/**
 * $Revision$
 * $Date$
 *
 * Copyright (C) 2007 Jive Software. All rights reserved.
 *
 * This software is published under the terms of the GNU Public License (GPL),
 * a copy of which is included in this distribution.
 */

package org.jivesoftware.openfire.gateway.avatars;

import org.xmpp.packet.JID;
import org.jivesoftware.util.NotFoundException;
import org.jivesoftware.util.Base64;
import org.jivesoftware.util.StringUtils;
import org.jivesoftware.database.DbConnectionManager;
import org.apache.log4j.Logger;

import java.util.Date;
import java.sql.PreparedStatement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * @author Daniel Henninger
 */
public class Avatar {

    static Logger Log = Logger.getLogger(Avatar.class);

    private static final String INSERT_AVATAR =
            "INSERT INTO gatewayAvatars(jid, xmppHash, legacyIdentifier, createDate, lastUpdate, imageData) " +
            "VALUES (?,?,?,?,?,?)";
    private static final String DELETE_AVATAR =
            "DELETE FROM gatewayAvatars WHERE jid=?";
    private static final String LOAD_AVATAR =
            "SELECT xmppHash, legacyIdentifier, createDate, lastUpdate " +
            "FROM gatewayAvatars WHERE jid=?";
    private static final String RETRIEVE_IMAGE =
            "SELECT imageData FROM gatewayAvatars WHERE jid=?";

    private JID jid;
    private String xmppHash;
    private String legacyIdentifier;
    private Date createDate;
    private Date lastUpdate;

    /**
     * Creates a new avatar entry.
     *
     * @param jid JID of the avatar.
     * @param legacyIdentifier Hash or whatever is necessary to identify the avatar on the legacy network.
     * @param imageData Base64 encoded image data.                 \
     * @throws IllegalArgumentException if any of the arguments are null.
     */
    public Avatar(JID jid, String legacyIdentifier, String imageData) throws IllegalArgumentException {
        if (jid == null || legacyIdentifier == null || imageData == null) {
            throw new IllegalArgumentException("Avatar: Passed null argument.");
        }
        this.jid = jid;
        this.legacyIdentifier = legacyIdentifier;
        try {
            MessageDigest md = MessageDigest.getInstance("SHA-1");
            md.update(Base64.decode(imageData));
            this.xmppHash = StringUtils.encodeHex(md.digest());
        }
        catch (NoSuchAlgorithmException e) {
            Log.error("Avatar: Unable to find support for SHA algorith?");
        }
        this.createDate = new Date();
        this.lastUpdate = new Date();
        try {
            insertIntoDb(imageData);
        }
        catch (SQLException e) {
            Log.error("Avatar: SQL exception while inserting avatar: ", e);
        }
    }

    /**
     * Creates a new avatar entry.
     *
     * @param jid JID of the avatar.
     * @param legacyIdentifier Hash or whatever is necessary to identify the avatar on the legacy network.
     * @param imageData Binary image data.
     * @throws IllegalArgumentException if any of the arguments are null.
     */
    public Avatar(JID jid, String legacyIdentifier, byte[] imageData) throws IllegalArgumentException {
        if (jid == null || legacyIdentifier == null || imageData == null) {
            throw new IllegalArgumentException("Avatar: Passed null argument.");
        }
        this.jid = jid;
        this.legacyIdentifier = legacyIdentifier;
        try {
            MessageDigest md = MessageDigest.getInstance("SHA-1");
            md.update(imageData);
            this.xmppHash = StringUtils.encodeHex(md.digest());
        }
        catch (NoSuchAlgorithmException e) {
            Log.error("Avatar: Unable to find support for SHA algorithm?");
        }
        this.createDate = new Date();
        this.lastUpdate = new Date();
        try {
            insertIntoDb(Base64.encodeBytes(imageData));
        }
        catch (SQLException e) {
            Log.error("Avatar: SQL exception while inserting avatar: ", e);
        }
    }

    /**
     * Loads an existing avatar.
     *
     * @param jid JID of the contact whose avatar we are retrieving.
     * @throws NotFoundException if avatar entry was not found in database.
     */
    public Avatar(JID jid) throws NotFoundException {
        this.jid = jid;
        loadFromDb();
        Log.debug("Loaded avatar for "+this.jid+" of hash "+this.xmppHash);
    }

    /**
     * Returns the JID of the avatar.
     *
     * @return JID of avatar.
     */
    public JID getJid() {
        return jid;
    }

    /**
     * Returns the XMPP hash (sha1).
     *
     * @return SHA1 based XMPP hash.
     */
    public String getXmppHash() {
        return xmppHash;
    }

    /**
     * Returns the legacy identifier of the avatar.
     *
     * This is completely up to the transport and is generally whatever type of hash is used on their end.
     *
     * @return Legacy identifier for the avatar.
     */
    public String getLegacyIdentifier() {
        return legacyIdentifier;
    }

    /**
     * Returns the creation date of the avatar in the database.
     *
     * @return Creation date of the avatar.
     */
    public Date getCreateDate() {
        return createDate;
    }

    /**
     * Returns the date the avatar was last updated.
     *
     * @return Last update date of the avatar.
     */
    public Date getLastUpdate() {
        return lastUpdate;
    }

    /**
     * Retrieves the actual image data for this avatar.
     *
     * @return The base64 encoded image data for the avatar.
     * @throws NotFoundException if avatar entry was not found in database.
     */
    public String getImageData() throws NotFoundException {
        Connection con = null;
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        String imageData = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(RETRIEVE_IMAGE);
            pstmt.setString(1, jid.toString());
            rs = pstmt.executeQuery();
            if (!rs.next()) {
                throw new NotFoundException("Avatar not found for " + jid);
            }
            imageData = rs.getString(1);
        }
        catch (SQLException sqle) {
            Log.error(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(rs, pstmt, con);
        }
        return imageData;
    }

    /**
     * Inserts a new avaar into the database.
     *
     * @param imageData Base64 encoded image data to be stored in database.
     * @throws SQLException if the SQL statement is wrong for whatever reason.
     */
    private void insertIntoDb(String imageData) throws SQLException {
        Connection con = null;
        PreparedStatement pstmt = null;
        boolean abortTransaction = false;
        try {
            con = DbConnectionManager.getTransactionConnection();
            pstmt = con.prepareStatement(DELETE_AVATAR);
            pstmt.setString(1, jid.toString());
            pstmt.executeUpdate();
            pstmt = con.prepareStatement(INSERT_AVATAR);
            pstmt.setString(1, jid.toString());
            pstmt.setString(2, xmppHash);
            pstmt.setString(3, legacyIdentifier);
            pstmt.setLong(4, createDate.getTime());
            pstmt.setLong(5, lastUpdate.getTime());
            pstmt.setString(6, imageData);
            pstmt.executeUpdate();
        }
        catch (SQLException sqle) {
            abortTransaction = true;
            throw sqle;
        }
        finally {
            DbConnectionManager.closeTransactionConnection(pstmt, con, abortTransaction);
        }
    }

    /**
     * Load avatar from database.
     *
     * @throws NotFoundException if avatar entry was not found in database.
     */
    private void loadFromDb() throws NotFoundException {
        Connection con = null;
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        try {
            con = DbConnectionManager.getConnection();
            pstmt = con.prepareStatement(LOAD_AVATAR);
            pstmt.setString(1, jid.toString());
            rs = pstmt.executeQuery();
            if (!rs.next()) {
                throw new NotFoundException("Avatar not found for " + jid);
            }
            this.xmppHash = rs.getString(1);
            this.legacyIdentifier = rs.getString(2);
            this.createDate = new Date(rs.getLong(3));
            this.lastUpdate = new Date(rs.getLong(4));
        }
        catch (SQLException sqle) {
            Log.error(sqle);
        }
        finally {
            DbConnectionManager.closeConnection(rs, pstmt, con);
        }
    }

}
